;;; inf-guile.el --- Emacs side of the "guile --emacs" protocol

;; Copyright (C) 2005, 2007 Thien-Thi Nguyen
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 3, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

;;; Author: Thien-Thi Nguyen <ttn@gnu.org>

;;; Commentary:

;; The Scheme module (ice-9 emacs) advertizes itself as "the Guile side
;; of the Emacs interface" but as of 2005-07-29 there doesn't seem to be
;; an Emacs Lisp side to it, so this file is an implementation of such,
;; reverse engineered (since protocol documentation is also lacking)
;; from the Scheme code.
;;
;; The user-visible aspects of the interface are (now) documented in
;; info node "(guile)REPL Under Emacs".

;;; Code:

(eval-when-compile (require 'cl))

(defvar inf-guile-transitions nil
  "Alist mapping transitions to their handlers.
Use `define-inf-guile-transition' to add to this.")

(defun define-inf-guile-transition (from to handler)
  (let* ((tag (cons from to))
         (bef (assoc tag inf-guile-transitions)))
    (if bef
        (setcdr bef handler)
      (push (cons tag handler) inf-guile-transitions))))

(define-inf-guile-transition t t
  (defun inf-guile-null-transition (proc old)
    "Return nil but otherwise do nothing useful.
The default transition handler specifies this function."
    nil))

(defun inf-guile-process-filter (proc string)
  (let* ((acc (let ((s (concat (process-get proc :output-so-far) string)))
                (process-put proc :output-so-far s)
                s))
         (cmd (string-match "\C-z[sfrFBx.SR]" acc))
         (raw (substring acc 0 cmd))
         (old (process-get proc :state))
         (new (case (and cmd (aref acc (1+ cmd)))
                ((?s) :enter-input-wait)
                ((?f) :exit-input-wait)
                ((?r) :enter-read-character)
                ((?F) :sending-error)
                ((?B) :sending-backtrace)
                ((?x) :sending-result)
                ((?.) :end-of-text)
                ((?S) :no-stack)
                ((?R) :no-source)
                (t old)))
         tag handler)
    (when cmd
      (process-put proc :state new)
      (process-put proc :output-so-far raw)
      (unless (eq old new)
        (setq handler (flet ((try (a d) (assoc (setq tag (cons a d))
                                               inf-guile-transitions)))
                        (cdr (or (try old new)
                                 (try old t)
                                 (try t new)
                                 (try t t)))))
        (if handler
            (funcall handler proc old)
          (error "No handler for %s -> %s transition!" old new)))
      (setq raw (substring acc (+ 2 cmd))))
    (process-put proc :output-so-far raw)
    (unless (string= "" raw)
      (inf-guile-process-filter proc ""))))

(defun inferior-guile ()
  "Start a \"guile --emacs\" subprocess.
Display the name of the new process object in
the echo area and return the new process object."
  (interactive)
  (let ((proc (start-process
               "inferior guile" nil "guile" "--emacs" "-c"
               (format "%s %s %s"
                       "(set-repl-prompt! #f)"
                       "(set! scm-repl-verbose #f)"
                       "(top-repl)"))))
    (process-put proc :state :hello)
    (set-process-filter proc 'inf-guile-process-filter)
    (message "Started: %S" (process-name proc))
    proc))

(provide 'inf-guile)

;;; inf-guile.el ends here
