;;; lex-mon.el --- monitor lexical analysis from Emacs

;; Copyright (C) 2003, 2007 Thien-Thi Nguyen

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 3, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

;;; Author: Thien-Thi Nguyen <ttn@gnu.org>

;;; Commentary:

;; This file provides the command `M-x lex-mon', which can be used
;; to visually monitor a lexical scan of a file (to debug the grammar
;; in most cases).  See docstring for `lex-mon' for more info.
;;
;; TODO: Rework to remove hardcoded /tmp/lex-mon-TMP file.
;;       Add option to recognize some tag and throw exception.
;;       Add region handling.
;;       Collect better stats, present them in a nicer way.

;;; Code:

(defvar lex-mon-subprocess-repl-return-count 0
  "Number of strings returned from the repl.")

(defun lex-mon-synchronous-send/return (proc message)
  "Return (TIME . FORM) where TIME is that returned by `current-time' and
FORM is the form read from the inferior lexer process."
  (put 'lex-mon 'sync-return "")
  (set-process-filter
   proc #'(lambda (proc string)
            (setq lex-mon-subprocess-repl-return-count
                  (1+ lex-mon-subprocess-repl-return-count))
            (let* ((so-far (get 'lex-mon 'sync-return))
                   (full   (put 'lex-mon 'sync-return (concat so-far string)))
                   (ans    nil))
              (when nil                 ; debugging
                (save-excursion
                  (set-buffer "***random***")
                  (goto-char (point-max))
                  (insert string)))
              (when (and
                     ;; Check for close-paren as a heuristic to avoid
                     ;; the more expensive `ignore-errors' machinery.
                     ;; Ideally, we would use a strchr workalike here.
                     (string-match ")" string)
                     (setq ans (ignore-errors (read full))))
                (put 'lex-mon 'sync-return
                     (cons (current-time) ans))))))
  (process-send-string proc message)
  (let (rv)
    ;; type change => break
    (while (stringp (setq rv (get 'lex-mon 'sync-return)))
      (accept-process-output proc))
    (put 'lex-mon 'sync-return "")
    (cdr rv)))

(defun lex-mon (grammar)
  "Enter a command loop to lexically scan the current buffer.
GRAMMAR names a lexical specification file to be passed to the
subprocess that actually does the scanning, which begins at the
current position and goes until the end of file.  Commands are
bound to one of these keys:

    SPC --- next token
    q   --- quit
    !   --- go until EOF without further interaction

As each token is recognized, Emacs displays its value in the echo
area.  After the `!' command, both `q' and `SPC' are disabled;
use `C-g' to break out of the loop."
  (interactive "fGrammar File: ")
  ;; All this lex-mon-TMP crap needs to be reworked (slightly).
  (when (get-buffer "lex-mon-TMP")
    (kill-buffer "lex-mon-TMP"))
  (let ((buf (get-buffer-create "lex-mon-TMP")))
    (append-to-buffer buf (point-min) (point-max))
    (put 'lex-mon 'orig-buffer-point (point))  ; ugh
    (switch-to-buffer buf)
    (goto-char (get 'lex-mon 'orig-buffer-point))
    (write-file "/tmp/lex-mon-TMP"))
  (let ((proc (start-process
               "lex-mon lexer" nil      ; "***random***"
               "guile-tools"
               "lexer-repl"
               "--emacs"
               "--offset" (number-to-string
                           (1- (get 'lex-mon 'orig-buffer-point)))
               "/tmp/lex-mon-TMP"
               (expand-file-name grammar)))
        (keep-going-p t)
        (prompt-head "(beginning of input)")
        (prompt-tail " SPC for next, ! to jam, or q to quit: ")
        (jammed nil)
        (token-count 0))
    (setq lex-mon-subprocess-repl-return-count 0)
    (while keep-going-p
      (let ((c (if jammed
                   (progn
                     ;; anti-alias measure: use explicit format string
                     (message "%s" prompt-head)
                     (sit-for 0)
                     jammed)
                 (read-char (concat prompt-head prompt-tail)))))
        (cond ((eq ?q c)
               (setq keep-going-p nil))
              ((or (eq 32 c) (eq ?! c))
               (when (eq ?! c) (setq jammed 32))
               (let ((ans (lex-mon-synchronous-send/return
                           proc "(next)\n")))
                 (setq token-count (1+ token-count))
                 (if (not (memq '<eoi> ans))
                     (let ((consumed (car ans)))
                       (forward-char consumed)
                       (setq prompt-head (format "%S" (cdr ans))))
                   (message "EOF (one more SPC to finish)")
                   (read-char)
                   (kill-process proc)
                   (setq keep-going-p nil))))
              (t
               (setq prompt-head "(sorry bad input)")))))
    (kill-buffer "lex-mon-TMP")
    (delete-file "/tmp/lex-mon-TMP")
    (message "stats: %d/%d (%f%%)"
             token-count lex-mon-subprocess-repl-return-count
             (/ (* 100.0 token-count) lex-mon-subprocess-repl-return-count))))

(provide 'lex-mon)

;;; lex-mon.el ends here
